# -*- coding: utf-8 -*-
from .. import py3
from ..logger import logging
from ..encoding import to_binary
from ..eastereggs.headers import get_headers

from cgi import parse_header, parse_multipart
from socket import error  # retry init'ing
from errno import EADDRINUSE  # retry init'ing
from time import sleep  # retry init'ing
from os import path

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)

try:
    if py3:
        # from socketserver import TCPServer
        from http.server import SimpleHTTPRequestHandler, HTTPServer
        from urllib.parse import parse_qs
    else:
        # from SocketServer import TCPServer
        from SimpleHTTPServer import SimpleHTTPRequestHandler
        from BaseHTTPServer import HTTPServer
        from urlparse import parse_qs
        # end if
except ImportError:
    raise  # try-except is only needed to satisfy PyCharm; to avoid PyCharm marking that as wrong import. As consequence the old error is reraised, as it really is an error.


# end try

class BetterHTTPRequestHandler(SimpleHTTPRequestHandler, object):
    def __init__(self, request, client_address, server):
        if py3:
            super(BetterHTTPRequestHandler, self).__init__(request, client_address, server)
        else:
            SimpleHTTPRequestHandler.__init__(self, request, client_address, server)
        self.static_files_dir = self.server.static_files_dir[:]  # force a copy

    def do_GET(self):
        logger.warn("do_GET is not overridden!")
        if self.path == "/":
            self.path = "/index.html"
        if False:  # Copy this and add your code here. ```if self.path == "/foo.bar":```
            pass
        else:
            self.path = self.static_files_dir + self.path  # e.g. localhost/123/foo.bar -> /path/to/script/webserver_files/123/foo.bar
            logger.debug("Requested file {file}".format(file=self.path))
            f = self.send_head()  # this handles 404'ing for us
            if f:
                try:
                    self.copyfile(f, self.wfile)
                finally:
                    f.close()
                    # end try
                    # end if
                    # end if-elif-else

    # end def

    def do_POST(self):
        logger.warn("do_POST is not overridden!")
        if self.path == "/":
            self.path = "/index.html"
        self.post_data = self.parse_POST()
        if False:  # Copy this and add your code here. ```if self.path == "/foo.bar":```
            pass
        else:
            self.path = self.static_files_dir + self.path  # e.g. localhost/123/foo.bar -> /path/to/script/webserver_files/123/foo.bar
            logger.debug("Requested file {file}".format(file=self.path))
            f = self.send_head()  # this handles 404'ing for us
            if f:
                try:
                    self.copyfile(f, self.wfile)
                finally:
                    f.close()
                # end try
            # end if
        # end if-elif-else
    # end def

    def parse_POST(self):
        ctype, pdict = parse_header(self.headers['content-type'])
        if "boundary" in pdict:
            pdict["boundary"] = to_binary(pdict["boundary"])
        if ctype == 'multipart/form-data':
            postvars = parse_multipart(self.rfile, pdict)
        elif ctype == 'application/x-www-form-urlencoded':
            length = int(self.headers['content-length'])
            postvars = parse_qs(
                self.rfile.read(length),
                keep_blank_values=1)
        else:
            postvars = {}
        return postvars

    # end def

    def write_text(self, msg, content_type="text/plain", is_binary=False):
        """
        Makes answering with text/data easier.

        :param msg: The text to send to the browser/client.
        :keyword content_type: If you don't like it to be text, change that here.
        :keyword is_binary: Text (unicode) needs to be converted to binary. If you already have binary (e.g. an PNG as binary data) you can set that here.
        :return: Nothing.
        """
        # Now do servery stuff.
        if not is_binary:
            msg = to_binary(msg)
        if msg is None:
            self.send_response(404)
            self.end_headers()
            return
        self.send_response(200)
        self.send_header("Content-type", content_type)
        self.send_header("Content-Length", str(len(msg)))
        for k, v in get_headers().items():
            self.send_header(k, v)
        self.send_header("X-Backend-created-by", "luckydonald")
        self.send_header("X-Licence", "Luna-Will-Cry-If-You-Modify-Or-Redistribute 1.0 or later")
        self.send_header("X-Licence-URL", "flutterb.at/lwc-1-0")
        self.end_headers()
        self.wfile.write(msg)
        return
    # end def

    def translate_path(self, path_):
        """
         Now accepts local fitting paths automatically
        E.g. "/path/to/www-dir/foo.png" is valid if that folder exists.
        Now it won't change the path to "/path/to/www-dir/foo.png/path/to/www-dir/foo.png", like it did before.
        :param path: path for the webserver.
        :return:
        """
        if path.exists(path_):
            return path_
        if py3:
            super(BetterHTTPRequestHandler, self).translate_path(path_)
        else:
            SimpleHTTPRequestHandler.translate_path(self, path_)
        return path_
    # end def

    def log_message(self, format, *args):
        logger.info(
            "%s - %s - [%s] %s\n" % (self.client_address[0], self.path, self.log_date_time_string(), format % args))

    def log_request(self, code='-', size='-'):
        logger.debug('%s - %s - "%s" %s %s', self.client_address[0], self.path, self.requestline, str(code), str(size))

    def log_error(self, format, *args):
        logger.error("%s - %s - %s" % (self.client_address[0], self.path, format % args))
    # end def
# end class


def start_a_webserver(handler, port, host="", static_files_dir=None):
    """
    Starts a ```TCPServer```, using the given ```handler```,
    :param handler: An handler instance, e.g. an ```BetterHTTPRequestHandler```
    :param port: The port where to serve on. For example ```80``` or ```8080``` for HTTP (```80``` often needs root privileges)
    :keyword host: Optional. A host where to serve on. If an empty string ```""``` (default) is given, all incoming connections are allowed. (from localhost, from lan, from internet, etc.)
    :keyword static_files_dir: Optional. Set the ```static_files_dir``` of an ```BetterHTTPRequestHandler``` handler.
    :return: The ```TCPServer``` created.
    """
    # assert (isinstance(handler, BetterHTTPRequestHandler))  # BaseRequestHandler?
    # handler = BetterHTTPRequestHandler()
    httpd = None
    started = False
    while not started:
        try:
            httpd = HTTPServer((host, port), handler)
            if issubclass(handler, BetterHTTPRequestHandler):
                if not static_files_dir:
                    logger.warn(
                        "`static_files_dir` kwarg-parameter should be used when a `BetterHTTPRequestHandler` is used.")
                httpd.static_files_dir = static_files_dir
            started = True
        except error as e:
            if e.errno in [EADDRINUSE]:
                logger.warn("Starting Server failed. Address already in use. Retrying.")
                sleep(1)
            else:
                raise
            # end if-else
        # end try
    # end while
    return httpd
# end def



# -*- coding: utf-8 -*-
from sys import version as python_version

__author__ = 'luckydonald'
__version__ = "0.69"  # had a leading zero before version "0.20" (i.e. "0.019").  PyPI always removed that anyway.
VERSION = __version__

py2 = True if python_version < '3' else False
py3 = not py2




from pip.commands.install import InstallCommand
from pip.baseparser import ConfigOptionParser
from pip.exceptions import PipError
from pip.utils import get_prog
import logging

logger = logging.getLogger(__name__)
install_cmd = InstallCommand()


def pip_install(package_name):
    parser_kw = {
        'usage': install_cmd.usage,
        'prog': '%s %s' % (get_prog(), install_cmd.name),
        'formatter': None,
        'add_help_option': False,
        'name': install_cmd.name,
        'description': install_cmd.__doc__,
        'isolated': False,
    }

    parser = ConfigOptionParser(**parser_kw)
    options, args = parser.parse_args(["install", package_name])
    try:
        status = install_cmd.run(options, args)
        if isinstance(status, int):
            logger.debug("Returned status code: {status}".format(status=status))
            return status == 0
    except (PipError, KeyboardInterrupt) as exc:
        logger.critical(str(exc))
        logger.debug('Exception information:', exc_info=True)
        return False
    except:
        logger.critical('Exception:', exc_info=True)
        return False
    return True



# -*- coding: utf-8 -*-
import getpass
import math

try:
    from .exceptions import assert_type_or_raise
except (ImportError, SystemError):
    # non-relative imports to enable doctests
    from luckydonaldUtils.exceptions import assert_type_or_raise
# end try

__author__ = 'luckydonald'

try:
    input = raw_input
except NameError:
    input = input
raw_input = input


def string_y_n(string, default=None):
    """
    Strict mapping of a given string to a boolean.

    If it is `'y'` or `'Y'`, `True` is returned.
    If it is `'n'` or `'N'`, `True` is returned.
    If it is empty or None (evaluates to False), and `default` is set, `default` is returned.
    Else a `ValueError` is raised.

    :param string: The input
    :type  string: str

    :param default: default result for empty input
    :type  default: None | bool

    :raises ValueError: If it is not any of ['y', 'Y', 'n', 'N']
    :return: result (True/False/default)
    :rtype:  bool
    """
    if not string and default is not None:
        return default
    if string not in ['y', 'Y', 'n', 'N']:
        raise ValueError('Please enter y or n.')
    if string == 'y' or string == 'Y':
        return True
    if string == 'n' or string == 'N':
        return False
        # end if


# end if


def string_is_yes(string, default=None):
    """
    Mapping of a given string to a boolean.

    If it is empty or None (evaluates to False), and `default` is set, `default` is returned.
    If the lowercase of the string is any of ['y', '1', 'yes', 'true', 'ja'], it will return `True`.
    Else it will return `False`

    :param string: The input
    :type  string: str

    :param default: default result for empty input
    :type  default: None | bool

    :return: result (True/False/default)
    :rtype:  bool
    """
    if not string and default is not None:
        return default
    # end if
    return string.lower() in ['y', '1', 'yes', 'true', 'ja']


# end if


# modified from http://code.activestate.com/recipes/541096-prompt-the-user-for-confirmation/
def confirm(prompt=None, default=False):
    """prompts for yes or no response from the user. Returns True for yes and
    False for no.

    'resp' should be set to the default value assumed by the caller when
    user simply presses ENTER.
    
    >>> confirm(prompt='Create Directory?', default=True)  # doctest: +SKIP
    Create Directory? [y]|n:
    True
    >>> confirm(prompt='Create Directory?', default=False)  # doctest: +SKIP
    Create Directory? [n]|y:
    False
    >>> confirm(prompt='Create Directory?', default=False)  # doctest: +SKIP
    Create Directory? [n]|y: y
    True

    """

    if prompt is None:
        prompt = 'Confirm'

    if default is not None:
        if default:
            prompt = '%s [%s]|%s: ' % (prompt, 'y', 'n')
        else:
            prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')
    else:
        prompt += ' y|n: '  # "%s y|n"
    # end if
    while True:
        ans = input(prompt)
        try:
            return string_y_n(ans)
        except ValueError as e:
            print(str(e))
            continue
    # end while
# end def


def answer(prompt=None, default=None):
    if prompt is None:
        prompt = 'Input'
    # end if
    if default is not None:
        prompt = '{prompt} [{default}]: '.format(prompt=prompt, default=default)
    else:
        prompt = '{prompt}: '.format(prompt=prompt)
    ans = None
    while not ans:
        ans = input(prompt)
        if not ans and default is not None:
            return default
            # end if
    # end while, when the user finally entered something
    return ans
# end def


def password(prompt=None, default=None):
    """
    If default is None, it will retry until the user inputs something.
    else set default="" to allow empty inputs.

    :param prompt: defaults to "Password"
    :param default:
    :return:
    """
    if prompt is None:
        prompt = 'Password'
    if default is not None:
        prompt = '{prompt} [{default}]: '.format(prompt=prompt, default=default)
    else:
        prompt = '{prompt}: '.format(prompt=prompt)
    passw = None
    while not passw:
        passw = getpass.getpass(prompt=prompt)
        if not passw and default is not None:
            return default
        # end if
    # end while
# end def




################################################
#                                              #
# SAFE EVAL    (could have holes in it...)     #
#                                              #
# modified from                                #
# http://lybniz2.sourceforge.net/safeeval.html #
#                                              #
################################################

# make a list of safe functions
eval_safe_builtin_list = ["math"]
eval_safe_builtin_function_list = {
    'acos': math.acos, 'asin': math.asin, 'atan': math.atan,
    'atan2': math.atan2, 'ceil': math.ceil, 'cos': math.cos, 'cosh': math.cosh,
    'degrees': math.degrees, 'e': math.e, 'exp': math.exp, 'fabs': math.fabs,
    'floor': math.floor, 'fmod': math.fmod, 'frexp': math.frexp, 'hypot': math.hypot,
    'ldexp': math.ldexp, 'log': math.log, 'log10': math.log10, 'modf': math.modf,
    'pi': math.pi, 'pow': math.pow, 'radians': math.radians, 'sin': math.sin,
    'sinh': math.sinh, 'sqrt': math.sqrt, 'tan': math.tan, 'tanh': math.tanh
}


class NotAllowed(Exception):
    """
    Thrown when a command in NoBuiltins is not allowed.
    """
    pass
# end class


class NoBuiltins(object):
    def __init__(self, allowed_builtins=eval_safe_builtin_list, allowed_functions=eval_safe_builtin_function_list,
                 allowed_vars=None):
        """
        :param allowed_builtins: List with names of functions.
        :type  allowed_builtins: list[str]
        :param allowed_functions: Dict with names of functions and the functions to be called.
        :type  allowed_functions: dict
        :param allowed_vars: Dict with allowed variables.
        :type  allowed_vars: dict

        """
        assert isinstance(allowed_builtins, list)
        self.allowed_builtins = tuple(allowed_builtins)  # tuples are not modifiable.
        if allowed_functions is None:
            allowed_functions = {}
        assert isinstance(allowed_functions, dict)
        self.allowed_functions = allowed_functions
        if allowed_vars is None:
            allowed_vars = {}
        assert isinstance(allowed_vars, dict)
        self.var_store = allowed_vars
    # end def

    def __getitem__(self, item):
        if item in self.var_store:
            return self.var_store[item]
        if item in self.allowed_functions:
            return self.allowed_functions[item]
        if item in self.allowed_builtins:
            return __builtins__[item]
        all = list(self.var_store.keys())
        all.extend(self.allowed_functions)
        raise NotAllowed(
            "{item} is not allowed, the supported commands are {allowed_funcs}, "
            "allowed variables are {allowed_vars}.".format(
                item=item, allowed_funcs=all, allowed_vars=self.var_store.keys()
            ))
    # end def
# end class


def safe_eval(user_input, allowed_values=NoBuiltins(eval_safe_builtin_list, eval_safe_builtin_function_list)):
    """
    Evals the `user_input` string.

    :param user_input:
    :param allowed_values: NoBuiltins object
    :type  allowed_values: NoBuiltins
    :param question:
    :return:
    :raises NotAllowed: The :class:`NoBuiltins` raises a :class:`NotAllowed` Exception when the command is not allowed.
    """
    assert_type_or_raise(user_input, str)
    assert isinstance(user_input, str)
    assert_type_or_raise(allowed_values, NoBuiltins)
    assert isinstance(allowed_values, NoBuiltins)
    result = eval(user_input, {"__builtins__": allowed_values}, allowed_values)
    return result


"""
>>> safe_eval("log10(12)")
1.0791812460476249
"""


from typing import Type, Dict, TypeVar, cast, Union, Optional, Callable

from ..exceptions import assert_type_or_raise
from ..dependencies import import_or_install

try:
    from pytgbot.api_types.receivable.updates import Update, Message
except ImportError:  # pragma nocover
    Update = import_or_install("pytgbot.api_types.receivable.updates", "Update")  # pip install pytgbot
    Message = import_or_install("pytgbot.api_types.receivable.updates", "Message")  # pip install pytgbot
# end try

__all__ = ['get_language_code', 'l_get']

LangClass = TypeVar('LangClass')


def get_language_code(update_msg_or_language_code: Union[Update, Message, str, None] = None) -> Optional[str]:
    assert_type_or_raise(update_msg_or_language_code, None, str, Message, Update, parameter_name="msg")

    # if is message, get the language_code from there.
    if isinstance(update_msg_or_language_code, Update):
        # we are in a update, we want any message
        if update_msg_or_language_code.message:
            update_msg_or_language_code = update_msg_or_language_code.message
        elif update_msg_or_language_code.callback_query:
            if update_msg_or_language_code.callback_query.from_peer and update_msg_or_language_code.callback_query.from_peer.language_code:
                return update_msg_or_language_code.callback_query.from_peer.language_code
            elif update_msg_or_language_code.callback_query.message:
                update_msg_or_language_code = update_msg_or_language_code.callback_query.message
            # end if
        # end if
    # end if
    if isinstance(update_msg_or_language_code, Message):
        # we are in a message, we want the language_code
        if update_msg_or_language_code.from_peer and update_msg_or_language_code.from_peer.language_code:
            return update_msg_or_language_code.from_peer.language_code
        elif update_msg_or_language_code.callback_query:
            return update_msg_or_language_code.callback_query.from_peer
        else:
            return None
        # end if
    # end if
    if not isinstance(update_msg_or_language_code, str):
        return None
    # end if
    return update_msg_or_language_code
# end def


def l_get(
        language_dict: Dict[str, LangClass], update_msg_or_language_code: Union[Update, Message, str, None] = None
) -> LangClass:
    """
    Retrieves a translation string for a given update/message/language_code.

    :param language_dict: the array containing the languages.
    :param update_msg_or_language_code: The stuff containing language info.
    :return: the fitting language class.
    """
    lang = get_language_code(update_msg_or_language_code)

    # if it is None (default, or getting from message failed), use the default
    if lang is None:
        return language_dict['default']
    # end if

    if lang in language_dict:
        return language_dict[lang]
    # end if

    # try splitting it "de_DE" => "de"
    part = lang.split('-')[0]
    if part in language_dict:
        return language_dict[part]
    # end of

    # try splitting it "de-DE" => "de"
    part = lang.split('_')[0]
    if part in language_dict:
        return language_dict[part]
    # end if

    # nothing did match, use the default
    return language_dict['default']
# end def



# -*- coding: utf-8 -*-
import re

__author__ = 'luckydonald'
__all__ = ['TELEGRAM_DOMAIN_REGEX', '_TELEGRAM_DOMAIN_REGEX', 'USER_LINK_REGEX', '_USER_LINK_REGEX']

_TELEGRAM_DOMAIN_REGEX = '(?:https?://)?(?:t(?:lgrm)?\.me|telegram\.(?:me|dog))/'
TELEGRAM_DOMAIN_REGEX = re.compile(_TELEGRAM_DOMAIN_REGEX)

from ..telegram import _USERNAME_REGEX

_USER_LINK_REGEX = '(?P<domain>' + _TELEGRAM_DOMAIN_REGEX + ')(?P<username>' + _USERNAME_REGEX + ')'
USER_LINK_REGEX = re.compile(_USER_LINK_REGEX)



# -*- coding: utf-8 -*-
"""
NOT PY2 ready!
>>> from luckydonaldUtils.eastereggs.quotes import get_quote
>>> get_quote("de")
>>> get_quote()
"""

from random import randint
from ..dependencies import import_or_install

try:
    from weakreflist.weakreflist import WeakList
except ImportError:  # pragma nocover
    WeakList = import_or_install("weakreflist.weakreflist.WeakList", "weakreflist")  # pip install weakreflist
# end try

__author__ = 'luckydonald'

QUOTES = [
    "Littlepip is best pony!",
    {"en": "May contain nuts", "de": "Kann Spuren von Nüssen enthalten"},
    {"en": "0% sugar", "de": "0% Zucker"},
    {"en": "100% pure", "de": "100% rein"},
    {"en": "12345 is a bad password", "de": "12345 ist ein schlechtes Passwort"},
    "90% bug free!",
    {"en": "Any computer is a laptop if you're brave enough",
     "de": "Jeder Computer ist ein Laptop, wenn du mutig bist"},
    "As seen on TV",
    "Ask your doctor",
    "Awesome",
    "20% cooler",
    "In 10 seconds flat",
    "I just don't know what went wrong...",
    "Liquid Pride",
    "Muffins!",
    '"[squeaks] [splatting] [magic zap] [magic zap] [sound of machinery] [magic zap]"',
    {"en": "Classy"},
    "Do not distribute",
    "Don't look directly at the bugs",
    "Free dental included",
    "Gasp!",
    {"de": "Keuch!"},
    "Google anlyticsed!",
    {"en": "Günter was here", "de": "Günter war hier"},
    "Hot tamale, hot hot tamale!",
    "Now in 3D",
    "Coded in Python 3",
    {"en": "Lives in a pineapple under the sea", "de": "Wohnt in 'ner Ananas ganz tief im Meer"},
    "Limited edition!",
    "Keep calm and carry on",
    {"de": "Nicht nur sauber sondern rein"},
    {"de": "Mit 99% weniger Zucker", "en": "With 99% less sugar"},
    {"de": "Besser als herkömmliche Produkte"},
    {"en": "Supercalifragilisticexpialidocious", "de": "Superkalifragilisticexpialigetisch"},
    "War. War never changes.",
    {"en": "The sky is the limit", "de": "Alles ist möglich"},
    {"en": "Turing complete", "de": "Turing-vollständig"},
    "Ultimate edition",
    {"en": "Uninflammable", "de": "Nicht brennbar"},
    {"en": "Water proof", "de": "Wasserdicht"},
    {"en": "Vote for net neutrality!", "de": "Stimme für die Netzneutralität!"},
    {"de": "Unplattbar"},
    "What's up, Doc?",
    {"en": "You can't explain that", "de": "Das kannst du nicht erklären"},
    "Multithreaded",
    "0xffffff",
    {"en": "#c0ffee is not a brown color", "de": "Die Farbe #c0ffee ist nicht braun"},
    {"en": "Cooler than Spock", "de": "Cooler als Spock"},
    {"de": "Achten Sie auf die Goldkante, es lohnt sich"},
    {"de": "Afri-Cola überwindet den toten Punkt"},
    {"de": "Dahinter steckt immer ein kluger Kopf"},
    {"de": "Mach mal Pause..."},
    {"de": "Dieses Produkt ist Glutenfrei"},
    {"de": "Milch macht müde Männer munter"},
    {"de": "Zu Risiken und Nebenwirkungen fragen Sie ihren Arzt oder Apotheker"},
    {"de": "Jetzt mit mehr Vitaminen"},
    {"de": "Hier könnte Ihre Werbung stehen"},
    "Pon <3",
    {"en": "Pon3 is best musician"},
    {"en": "Batteries not included", "de": "Batterien nicht enthalten"},
    {"de": "Abbildung ähnlich"},
    {"de": "Serviervorschlag"},
    {"en": "So kawaii"},
    "Aloha",
    {"de": "Anscheinend können Sie lesen", "en": "It seems you can read"},
    "DINKELBERG!",
    # {"de": ".sträwkcür tsi txeT reseiD", "en": ".desrever si txet sihT"},
    "The cake is a lie",
    # {"de": ""},
    # {"en": "", "de": ""},
]
QUOTES_BY_LANGUAGE = {"all": WeakList()}
__did_init = [False]  # list is an access hack


def init_quotes():
    """
    Store them ordered.

    Will be called on module load.
    """
    # reset QUOTES_BY_LANGUAGE
    for key in list(QUOTES_BY_LANGUAGE):
        del QUOTES_BY_LANGUAGE[key]
    QUOTES_BY_LANGUAGE["all"] = WeakList()
    # fill QUOTES_BY_LANGUAGE, with language first.
    for quote in QUOTES:
        if isinstance(quote, dict):  # specific language(s)
            for language, text in quote.items():  # todo: python 2 iter_something()
                if not language in QUOTES_BY_LANGUAGE:
                    QUOTES_BY_LANGUAGE[language] = WeakList()
                # end if
                QUOTES_BY_LANGUAGE[language].append(quote[language])
                # end for
        elif isinstance(quote, str):
            QUOTES_BY_LANGUAGE["all"].append(quote)
            # end if
    # end for
    __did_init[0] = True  # list is an access hack


# end def

def get_quote(language="en"):
    """
    Get a random quote.

    :param language: default: "en"
    :return:
    """
    if not __did_init[0]:
        init_quotes()
    quotes_language = QUOTES_BY_LANGUAGE[language]
    quotes_universal = QUOTES_BY_LANGUAGE["all"]
    length_universal = len(quotes_universal)
    index = randint(0, (length_universal + len(quotes_language)) - 1)
    if index < length_universal:
        return quotes_universal[index]
    else:
        return quotes_language[index - length_universal]
        # end if


# end def

init_quotes()


class Quote(object):
    """
    Unused by now.
    """
    text = None
    url = None

    def __init__(self, value, url):
        self.text = value
        self.url = url

    def __str__(self):
        return self.text
        # end def __str__
        # end class Quote



import functools
import inspect
import logging
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)

__author__ = 'luckydonald'
__all__ = ["caller", "deprecated", "gone", "cached"]


def caller(func):
    """
    functions decorated with this will be called with an `call` kwarg, containing information about the function itself, and the caller.
    If the caller could not be fetched correctly, the `caller`s attributes all will be `None`.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        try:
            stack = inspect.stack()[1]
            call = {"self": {"name": func.__name__, "file": func.__code__.co_filename,
                             "line": func.__code__.co_firstlineno + 1},
                    "caller": {"name": stack[3], "file": stack[1], "line": stack[2], "code": stack[4][0]}
                    }
        except TypeError:
            call = {"self": {"name": func.__name__, "file": func.__code__.co_filename,
                             "line": func.__code__.co_firstlineno + 1},
                    "caller": {"name": None, "file": None, "line": None, "code": None}
                    }
        kwargs["call"] = call
        return func(*args, **kwargs)

    return new_func


def deprecated(func, message=None):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being logged
    when the function is used."""

    if isinstance(func, str):  # @deprecated("message")
        message = func

        def inners(func):  # @deprecated("message")(func)
            return deprecated(func, message)  # @deprecated("message")(func) ->  @deprecated(func, "message")

        return inners

    # end if
    # -> @deprecated(func)

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        try:
            try:  # python 2 + python 3
                f_name = func.__name__
                f_file = func.__code__.co_filename
                f_line = func.__code__.co_firstlineno + 1
            except AttributeError:  # maybe some old python 2?
                f_name = func.func_code.co_name
                f_file = func.func_code.co_filename
                f_line = func.func_code.co_firstlineno + 1
            # end try
            try:
                logger.warning(
                    "Call to deprecated function {f_name} at file {f_file}:{f_line}\n"
                    "{msg}Called from function {c_name} at file {c_file}:{c_line}\nLine: {c_code}".format(
                        f_name=f_name, f_file=f_file,
                        f_line=f_line,
                        msg="Warning: " + message.strip() + "\n" if message else "",
                        c_name=inspect.stack()[1][3], c_file=inspect.stack()[1][1],
                        c_line=inspect.stack()[1][2], c_code=inspect.stack()[1][4][0],
                    ))
            except TypeError:
                logger.warning(
                    "Call to deprecated function {f_name} at file {f_file}:{f_line}{msg}".format(
                        f_name=f_name, f_file=f_file,
                        f_line=f_line, msg="\nWarning: " + message.strip() if message else ""
                    ))
        except (AttributeError, TypeError):
            logger.warning(
                "Call to deprecated function.{msg}".format(msg="\nWarning: " + message.strip() if message else ""),
                exc_info=True
            )
        return func(*args, **kwargs)

    return new_func  # @deprecated(func)


# end def
"""
Test would be
>> @deprecated
>> def foo(): pass

>> @deprecated("Custom Message")
>> def foo(): pass
"""


def gone(func):
    """This is a decorator which can be used to mark functions
    as deprecated, and gone. It will raise a NotImplementedError
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        raise NotImplementedError(
            "Call to deprecated function {} at file {}:{}. Called from function {} at file {}:{}\nLine: {}".format(
                func.__name__, func.__code__.co_filename,  # python2: func.func_code.co_filename
                func.__code__.co_firstlineno + 1,  # python2: func.func_code.co_firstlineno
                inspect.stack()[1][3], inspect.stack()[1][1], inspect.stack()[1][2], inspect.stack()[1][4][0]
            )
        )

    # end def
    return new_func
# end def


def cached(max_age=None):
    """
    Cache a function.
    You can specify a max_age (:class:`datetime.timedelta`) after when the function will be called again.

    :keyword max_age: timedelta of how long we should keep the cache. None means forever.
    :type    max_age: None or timedelta
    """
    # http://stackoverflow.com/a/30698822/3423324
    # _format = {"returns": "I am a return value!!", "max_age": timedelta(seconds=60), "last_hit": datetime.now()}
    memo = {}
    import json

    def stringify(obj):
        if isinstance(obj, (list, tuple)):
            return repr([stringify(x) for x in obj])
        if isinstance(obj, dict):
            res = {}
            for k, v in obj.items():
                res[k] = stringify(k)
            # end for
            return repr(res)
        try:
            return json.dumps(obj)
        except TypeError:
            return str(repr(obj)) + str(obj) + str(id(obj))
            # end if

    def func_wrapper(function, **decorator_kwargs):
        def wrapper(*args, **kwargs):
            args_key = stringify([args, kwargs])
            now = datetime.now()
            if "max_age" in decorator_kwargs:
                max_age_ = decorator_kwargs["max_age"]
            else:
                max_age_ = max_age
            if (args_key) in memo.keys():
                logger.debug(memo[args_key])
                if not max_age_ or now - memo[args_key]["last_hit"] < memo[args_key]["max_age"]:
                    return memo[args_key]["returns"]
                    # end if
            # end if
            res = function(*args, **kwargs)
            memo[args_key] = {"returns": res, "last_hit": now, "max_age": max_age_}
            return res

        return wrapper

    try:
        is_callable = callable(max_age)
    except (SyntaxError, NameError):
        is_callable = inspect.isfunction(max_age) or inspect.isbuiltin(max_age)
    # end try
    if is_callable:  # max_age was not provided, max_age is actually our function to decorate
        func = max_age
        return func_wrapper(func, max_age=None)
    return func_wrapper  # else: max_age is max_age, we need to return a decorator accepting the function
# end def



# -*- coding: utf-8 -*-
import re

__author__ = 'luckydonald'
__all__ = [
    'USERNAME_REGEX', '_USERNAME_REGEX', 'USER_AT_REGEX', '_USER_AT_REGEX',
    'FULL_USERNAME_REGEX', '_FULL_USERNAME_REGEX'
]

_USERNAME_REGEX = '[a-zA-Z](?:[a-zA-Z0-9]|_(?!_)){3,30}[a-zA-Z0-9]'  # https://regex101.com/r/nZdOHS/2
USERNAME_REGEX = re.compile(_USERNAME_REGEX)

_USER_AT_REGEX = '@(?P<username>' + _USERNAME_REGEX + ')'
USER_AT_REGEX = re.compile(_USER_AT_REGEX)

from .urls.telegram import _TELEGRAM_DOMAIN_REGEX
_FULL_USERNAME_REGEX = '(?P<prefix>(?P<domain>' + _TELEGRAM_DOMAIN_REGEX + ')|@)(?P<username>' + _USERNAME_REGEX + ')'
FULL_USERNAME_REGEX = re.compile(_FULL_USERNAME_REGEX)

# backwards compatibility:
from .urls.telegram import TELEGRAM_DOMAIN_REGEX, _USER_LINK_REGEX, USER_LINK_REGEX
